/* 
   This is the graphical X-Windows front-end for NetBrisk.
    Copyright (C) 1995 Brendan Bartlett

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Toggle.h>
#include <X11/xpm.h>

#include "brisk.h"

extern int server_fd, numplayers, play_type, my_number;
extern char *errormsg;
 
static Widget      toplevel, parentbox, box[4], tablew;                 /* These are the parent widgets */
static Widget      hand[3], table[5];                            /* These contain the graphics */
static Widget      status, scorelabel, numberlabel;                          /* text labels */
static Widget      winnerlabel, winnertoggle;

char *ourname;

Pixmap *back, *blank;
Pixmap *images[4][10];

extern void process_input(void);
extern void open_server(char *, int);
extern void ok_score(void);

void main(int, char **);
void hand1(Widget, XtPointer, XtPointer);
void hand2(Widget, XtPointer, XtPointer);
void hand3(Widget, XtPointer, XtPointer);
void show_brisk(int, int);
void show_card(int, int, int);
void show_table(int, int, int);
void show_clear_table(char, char, int);
void show_move_expected(int);

static void SetUpScreen(void);
static int ReadBitmaps(void);


void main(int argc, char **argv)
{
  Arg parentargs[] = {
    {XtNwidth, (XtArgVal) 800},
    {XtNheight, (XtArgVal) 600},
    {XtNborderWidth, (XtArgVal) 5},
  };
  int i, port = 0;
  XEvent event;
  XtPointer condition;
  char buf[80];

  ourname = argv[0];

  if(argc == 1)
    {
      printf("usage: %s <hostname> <port>\n", argv[0]);
      printf(" hostname is the name where the SERVER is running.\n");
      printf(" port is an optional setting to connect to the server.\n");
      exit(0);
    }

  if(argc == 3)
    port = atoi(argv[2]);

  toplevel = XtInitialize(ourname, "XBrisk", NULL, 0, &argc, argv);

  parentbox = XtCreateManagedWidget(ourname, viewportWidgetClass, toplevel, parentargs, XtNumber(parentargs));

  SetUpScreen();

  if(!ReadBitmaps())
    exit(0);

  
  open_server(argv[1], port);

  sprintf(buf, "Player %d", my_number);
  XtVaSetValues(numberlabel, XtNlabel, buf, NULL);

  condition = (XtPointer) XtInputReadMask;

  XtAddInput(server_fd, condition, (XtInputCallbackProc) process_input, NULL);
  XtRealizeWidget(toplevel);


  XtMainLoop();
}

void show_brisk(int suit, int weight)
{
  if(suit == -1 && weight == -1)
    XtVaSetValues(table[0], XtNbitmap, blank, NULL);    
  else
    XtVaSetValues(table[0], XtNbitmap, images[suit][weight], NULL);
}

void show_card(int cardnum, int suit, int weight)
{
/*
  static int otherdown = 0;
*/

  if(suit == -1 && weight == -1)
      XtVaSetValues(hand[cardnum], XtNbitmap, blank, NULL);
  else
    XtVaSetValues(hand[cardnum], XtNbitmap, images[suit][weight], NULL);
}

void show_table(int tablepos, int suit, int weight)
{
  XtVaSetValues(table[tablepos + 1], XtNbitmap, images[suit][weight], NULL);
}

/* This is called at the end of every hand.
 * It will display a dialog box with the winner, update the scores on screen
 * and finally clear the table of all cards.
 */
void show_clear_table(char myscore, char otherscore, int lastwin)
{
  int i;
  char buf[80], buf2[80];

  if(numplayers == 2 || ( (numplayers == 4) && (play_type == TEAM) ) )
    {
      sprintf(buf, "Scores: %d for you, %d for other %s.", myscore, otherscore, (numplayers == 2) ? "player" : "team");
      XtVaSetValues(scorelabel, XtNlabel, buf, NULL);
    }
  
  if(lastwin == my_number)
    sprintf(buf2, "You won the last hand!");
  else
    sprintf(buf2, "Player %d won the last hand.", lastwin);
  
  XtVaSetValues(winnerlabel, XtNlabel, buf2, XtNsensitive, True, NULL);
  XtVaSetValues(winnertoggle, XtNsensitive, True, NULL);
}

void clear_table()
{
  int i;

  ok_score();

  XtVaSetValues(winnerlabel, XtNsensitive, False, NULL);
  XtVaSetValues(winnertoggle, XtNsensitive, False, XtNstate, False, NULL);

  for(i = 1; i < 5; i++)
    XtVaSetValues(table[i], XtNbitmap, blank, NULL);
}

void show_move_expected(int player)
{
  char buf[80];
  
  if(player == my_number)
    sprintf(buf, "It's your turn!");
  else
    sprintf(buf, "It's player %ds turn\n", player);
  
  XtVaSetValues(status, XtNlabel, buf, NULL);
}

void hand1(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtVaSetValues(w, XtNstate, False, NULL);

  if(!play_card(0))
    {
      XtVaSetValues(status, XtNlabel, errormsg, NULL);
      return;
    }

  XtVaSetValues(hand[0], XtNbitmap, blank, NULL);
  XtVaSetValues(status, XtNlabel, " ", NULL);
}

void hand2(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtVaSetValues(w, XtNstate, False, NULL);
  
  if(!play_card(1))
    {
      XtVaSetValues(status, XtNlabel, errormsg, NULL);
      return;
    }

  XtVaSetValues(hand[1], XtNbitmap, blank, NULL);
  XtVaSetValues(status, XtNlabel, " ", NULL);
}

void hand3(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtVaSetValues(w, XtNstate, False, NULL);

  if(!play_card(2))
    {
      XtVaSetValues(status, XtNlabel, errormsg, NULL);
      return;
    }

  XtVaSetValues(hand[2], XtNbitmap, blank, NULL);
  XtVaSetValues(status, XtNlabel, " ", NULL);
}


/* Setup the screen.  Create the widgets for the boxes
 * for the various regions of the screen.
 *
 * No return
 */
static void SetUpScreen(void)
{
  int xoff[] = { 0, 458, 0, 458 };
  int yoff[] = { 0, 0, 435, 435 };

  int i;
  Arg arguments[] = {                         /* Don't change the order!! */
    { XtNwidth, (XtArgVal) 360 },
    { XtNheight, (XtArgVal) 150 },
    { XtNborderWidth, (XtArgVal) 0 },
    { XtNx, (XtArgVal) 0 },
    { XtNy, (XtArgVal) 0 }
  };

  for(i = 0; i < 4; i++)
    {
      arguments[3].value = xoff[i];
      arguments[4].value = yoff[i];

      box[i] = XtCreateManagedWidget(ourname, boxWidgetClass, parentbox, arguments, XtNumber(arguments));
    }
  
  /* This is the table for 4 players */

  arguments[0].value = 580;
  arguments[1].value = 150;
  arguments[2].value = 1;
  arguments[3].value = 120;
  arguments[4].value = 225;
  
  tablew = XtCreateManagedWidget(ourname, boxWidgetClass, parentbox, arguments, XtNumber(arguments));

  status = XtCreateManagedWidget(ourname, labelWidgetClass, parentbox, NULL, 0);
  XtVaSetValues(status, XtNjustify, XtJustifyLeft, XtNx, 5, XtNy, 170, XtNwidth, 160, XtNlabel, " ", NULL);
  
  scorelabel = XtCreateManagedWidget(ourname, labelWidgetClass, parentbox, NULL, 0);
  XtVaSetValues(scorelabel, XtNjustify, XtJustifyLeft, XtNx, 5, XtNy, 200, XtNwidth, 300, XtNlabel, " ", NULL);

  winnerlabel = XtCreateManagedWidget(ourname, labelWidgetClass, parentbox, NULL, 0);
  XtVaSetValues(winnerlabel, XtNx, 320, XtNy, 170, XtNjustify, XtJustifyLeft, XtNwidth, 200, XtNsensitive, True, 
		XtNlabel, " ", NULL);

  numberlabel = XtCreateManagedWidget(ourname, labelWidgetClass, parentbox, NULL, 0);
  XtVaSetValues(numberlabel, XtNx, 365, XtNy, 75, XtNjustify, XtJustifyLeft, XtNwidth, 100, XtNsensitive, True, 
		XtNlabel, " ", NULL);

  winnertoggle = XtCreateManagedWidget(ourname, toggleWidgetClass, parentbox, NULL, 0);
  XtVaSetValues(winnertoggle, XtNx, 320, XtNy, 190, XtNjustify, XtJustifyLeft, XtNsensitive, False, XtNlabel, " Ok ", NULL);
  XtAddCallback(winnertoggle, XtNcallback, clear_table, NULL);



}

static int ReadBitmaps(void)
{
  Widget test;
  Arg arguments[] = {
    { XtNbitmap, (XtArgVal) back },
    { XtNx, (XtArgVal) 0 },
    { XtNy, (XtArgVal) 0 }
  };
  void *callbacks[] = { hand1, hand2, hand3 };
  char *bitnames[] = {"cup", "sun", "bone", "sword"};
  char fname[20];
  int i, j, suit, weight;
  

  if((XpmReadFileToPixmap(XtDisplay(toplevel), DefaultRootWindow(XtDisplay(toplevel)), "./bitmaps/back.xpm", (Pixmap *)&back, NULL, NULL) == -1))
    {
      printf("Fatal: Error trying to read ./bitmaps/back.xpm bitmap\n");
      return 0;
    }
  
  if(XpmReadFileToPixmap(XtDisplay(toplevel), DefaultRootWindow(XtDisplay(toplevel)), "./bitmaps/blank.xpm", (Pixmap *)&blank, NULL, NULL) == -1)
    {
      printf("Fatal: Error trying to read ./bitmaps/blank.xpm bitmap\n");
      return 0;
    }
  
  /* Read in ALL of the bitmaps */


  for(suit = 0; suit < 4; suit++)
    {
      fprintf(stderr, "Reading suit %s", bitnames[suit]);

      for(weight = 0; weight < 10; weight++)
	{
	  sprintf(fname, "./bitmaps/%s%d.xpm", bitnames[suit], weight);
	  if(XpmReadFileToPixmap(XtDisplay(toplevel), DefaultRootWindow(XtDisplay(toplevel)), fname, (Pixmap *)&images[suit][weight], NULL, NULL) == -1)
	    {
	      printf("Fatal: Error trying to read %s bitmap\n", fname);
	      return 0;
	    }
	  fprintf(stderr, ".");
	}
      fprintf(stderr,"\n");
    }


  /* Create our hand */
  arguments[0].value = (XtArgVal) blank;

  for(i = 0; i < 3; i++)
    {
      hand[i] = XtCreateManagedWidget(ourname, toggleWidgetClass, box[0], arguments, XtNumber(arguments));
      XtAddCallback(hand[i], XtNcallback, callbacks[i], NULL);
      arguments[1].value += 120;
    }

  /* Make the other player's cards the back card */
  arguments[0].value = (XtArgVal) back;
  arguments[1].value = (XtArgVal) 0;

  for(i = 1; i < 4; i++)
    {
      test = XtCreateManagedWidget(ourname, labelWidgetClass, box[i], arguments, XtNumber(arguments));
      arguments[1].value += 120;
      test = XtCreateManagedWidget(ourname, labelWidgetClass, box[i], arguments, XtNumber(arguments));
      arguments[1].value += 120;
      test = XtCreateManagedWidget(ourname, labelWidgetClass, box[i], arguments, XtNumber(arguments));
    }

  arguments[0].value = (XtArgVal) blank;
  arguments[1].value = 0;

  for(i = 0; i < 5; i++)
    {
      table[i] = XtCreateManagedWidget(ourname, labelWidgetClass, tablew, arguments, XtNumber(arguments));
      arguments[1].value += 120;
    }

/*
  arguments[0].value = (XtArgVal) images[0][0];
  XtSetValues(test, arguments, XtNumber(arguments));
*/

  return 1;
}




